/* ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is ThumbsDown.
 *
 * The Initial Developer of the Original Code is
 * Atte Kemppilä.
 * Portions created by the Initial Developer are Copyright (C) 2008
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 * 
 * ***** END LICENSE BLOCK ***** */


// FIXME: the whole wizard. split this into one xul/js per page. redesign "model"


const Cc = Components.classes;
const Ci = Components.interfaces;
const Cr = Components.results;

const atomSvc = Cc["@mozilla.org/atom-service;1"].getService(Ci.nsIAtomService);
const ioSvc = Cc["@mozilla.org/network/io-service;1"].getService(Ci.nsIIOService);
const scriptMgr = Cc["@thumbsdown.mozdev.org/script-manager;1"].getService(Ci.nsISupports).wrappedJSObject;

const TARGET_IMAGE = 1;
const TARGET_PAGE  = 2;

const MODE_CREATE = 1;
const MODE_MODIFY = 2;

const TYPE_DIRECT = 0;
const TYPE_REGEXP = 1;
const TYPE_CUSTOM = 2;

const TYPE_REGEXP_THUMBNAIL = 0;
const TYPE_REGEXP_LINK      = 1;

var wizard = null;
var pages = {};
var page = null;

var isAdvancing = false;


function init()
{
    wizard = $("thumbsdown_scriptWizard");
    stringBundle = $("stringbundle");
    model.init();
}

function onUnload()
{
    for each (thumbnail in model.thumbnails)
    {
        if (thumbnail.targetRequest.timeoutId)
        {
            window.clearTimeout(thumbnail.targetRequest.timeoutId);
            delete thumbnail.targetRequest.timeoutId;
        }
        thumbnail.targetRequest.abort();
    }
}

function onPageShow(e)
{
    var pageId = e.target.pageid;
    page = pages[pageId];
    if (page.onPageShow)
        page.onPageShow();
    updateWizardButtons();
}

function onPageAdvanced()
{
    if (isAdvancing)
        isAdvancing = false;
    else if (page.onPageAdvanced)
        return page.onPageAdvanced();
    return true;
}

function onPageRewound()
{
    if (page.onPageRewound)
        return page.onPageRewound();
    return true;
}

function advance(pageId)
{
    isAdvancing = true;
    wizard.advance(pageId);
    return false;
}

function updateWizardButtons()
{
    wizard.canAdvance = (page.canAdvance ? page.canAdvance() : true);
    wizard.canRewind  = (page.canRewind  ? page.canRewind()  : true);
}


var model =
{
    get thumbnails() { return this._thumbnails; },
    set thumbnails(value)
    { 
        this._thumbnails = value;
        this.selected = null;
    },
    
    get selected() { return this._selected; },
    set selected(value)
    { 
        this._selected = value;
        this.target = null;
        this.match = null;
        this.linkUrlTokens = null;
        this.mode = null;
        this.images = null;
    },
    
    get mode() { return this._mode; },
    set mode(value)
    {
        if (value == this._mode)
            return;
        
        this._mode = value;
        this.scriptId = null;
    },
    
    get isCreate() { return (this.mode == MODE_CREATE); },
    get isModify() { return (this.mode == MODE_MODIFY); },
    
    get scriptId() { return this._scriptId; },
    set scriptId(value)
    {
        if (value == this._scriptId)
            return;
        
        this._scriptId = value;
        this.name0 = null;
        this.matcher0 = null;
        this.type = null;
        this.patternSource = null;
        this.pattern = null;
        this.template = null;
    },
    
    get script() { return scriptMgr.getScript(this.scriptId); },
    
    get name0() { return this._name0; },
    set name0(value)
    {
        this._name0 = value;
        this.name = null;
    },
    
    get matcher0() { return this._matcher0; },
    set matcher0(value)
    {
        this._matcher0 = value;
        this.matcher = null;
    },
    
    get isDirect() { return (this.type == TYPE_DIRECT); },
    get isRegExp() { return (this.type == TYPE_REGEXP); },
    get isCustom() { return (this.type == TYPE_CUSTOM); },
    
    get images() { return this._images; },
    set images(value)
    {
        this._images = value;
        this.patternSource = null;
        this.pattern = null;
        this.template = null;
    },
    
    isInitialized: false,
    
    init: function()
    {
        if (this.isInitialized)
            return;
        this.isInitialized = true;
        
        this.thumbnails = [];
        var thumbnails = this.thumbnails;
        var doc = window.arguments[0];
        
        if (! doc.images)
            return;
        
        for (var i = 0; i < doc.images.length; i++)
        {
            try
            {
                var image = doc.images[i];
                var anchor = image.parentNode;
                if (anchor && anchor.href && (anchor instanceof HTMLAnchorElement))
                {
                    var thumbnail =
                    {
                        image: image,
                        anchor: anchor,
                        thumbnailUrl: ioSvc.newURI(image.src, null, null).QueryInterface(Ci.nsIURL),
                        linkUrl: ioSvc.newURI(anchor.href, null, null).QueryInterface(Ci.nsIURL),
                        index: thumbnails.length
                    }
                    thumbnail.targetRequest = this._startTargetRequest(thumbnail);
                    thumbnails.push(thumbnail);
                }
            }
            catch (e) { /* ignore */ }
        }
        
        // TODO: keep name list in scriptMgr
        this.names = scriptMgr.sqlExec.fetchAll("getNames");
    },
    
    _startTargetRequest: function(thumbnail)
    {
        function onThumbnailTargetUpdate(thumbnail)
        {
            if (page && page.onThumbnailTargetUpdate)
                page.onThumbnailTargetUpdate(thumbnail);
        }
        
        function onError(e)
        {
            if (thumbnail.target)
                return;
            
            var request = e.target;
            
            if (request.timeoutId)
            {
                window.clearTimeout(request.timeoutId);
                delete request.timeoutId;
            }
            
            thumbnail.target = {success: false, type: "error"};
            
            /*
             * If request fails and wizard page has been changed already, you get:
             * 0x80040111 (NS_ERROR_NOT_AVAILABLE) [nsIXMLHttpRequest.status]
             * https://bugzilla.mozilla.org/show_bug.cgi?id=301705
             */
            try { thumbnail.target.status = request.status; }
            catch (e if (e.result == Cr.NS_ERROR_NOT_AVAILABLE)) { /* ignore */ }
            
            onThumbnailTargetUpdate(thumbnail);
        }
        
        function onLoad(e)
        {
            if (thumbnail.target)
                return;
            
            var request = e.target;
            
            if (request.status != 200)
            {
                onError(e);
                return;
            }
            
            if (request.timeoutId)
            {
                window.clearTimeout(request.timeoutId);
                delete request.timeoutId;
            }
            
            thumbnail.target = {success: true, type: "unknown", status: request.status};
            var contentType = thumbnail.contentType = request.getResponseHeader("Content-Type");
            if (contentType)
            {
                var match =  /^([-\w]+\/[-\w]+)/.exec(contentType);
                if (match)
                {
                    var mimeType = thumbnail.target.mimeType = match[1];
                    if (/^image\/\w+/.test(mimeType))
                        thumbnail.target.type = "image";
                    else if (mimeType == "text/html")
                        thumbnail.target.type = "page";
                }
            }
            onThumbnailTargetUpdate(thumbnail);
        }
        
        function onTimeout(request)
        {
            request.abort();
            delete request.timeoutId;
            
            if (thumbnail.target)
                return;
            
            thumbnail.target = {success: false, type: "timeout"};
            onThumbnailTargetUpdate(thumbnail);
        }
        
        var request = new XMLHttpRequest();
        request.open("HEAD", thumbnail.anchor.href, true);
        request.setRequestHeader("Referer", window.arguments[0].location.href);
        request.onload = onLoad;
        request.onerror = onError;
        request.send(null);
        
        var timeout = nsPreferences.getIntPref("extensions.thumbsdown.scriptwizard.requestTimeout");
        request.timeoutId = window.setTimeout(function() { onTimeout(request); }, timeout * 1000);
        
        return request;
    }
}


pages.thumbnails = 
{
    isInitialized: false,
    tree: null,
    thumbnailIndex: null,
    
    get targetCount()
    {
        if (! this._targetCount)
        {
            this._targetCount = {image: 0, page: 0, unknown: 0, error: 0};
            for each (thumbnail in model.selected)
                if (thumbnail.target)
                    this._targetCount[thumbnail.target.type]++;
        }
        return this._targetCount;
    },
    
    onPageShow: function()
    {
        if (! this.isInitialized)
        {
            init();
            
            this.isInitialized = true;
            this.tree = $("tree_thumbnails");
            this.tree.view = new ThumbnailTreeView(model.thumbnails);
            
            this._updateSelected();
        }
    },
    
    onPageAdvanced: function()
    {
        if (model.match)
            return advance("script");
        
        if (! model.linkUrlTokens)
        {
            // try to find an existing script that matches all the select thumbnails
            var matchesArray = [];
            for each (thumbnail in model.selected)
            {
                var matches = scriptMgr.lookupAll(thumbnail.anchor.href, thumbnail.image.src);
                if (matches.length == 0)
                    break;
                matchesArray.push(matches);
            }
            
            if (matchesArray.length == model.selected.length)
            {
                for (var i = 0; i < matchesArray[0].length; i++)
                {
                    var script = matchesArray[0][i];
                    var isSameScript = function(s) { return (s.id == script.id); };
                    for (var k = 1; k < matchesArray.length; k++)
                    {
                        var matches = matchesArray[k];
                        if (! matches.some(isSameScript))
                            break;
                    }
                    
                    if (k == matchesArray.length)
                    {
                        // matching script has been found
                        model.match = script;
                        return advance("script");
                    }
                }
            }
            
            var linkUrls = [thumbnail.linkUrl for each (thumbnail in model.selected)];
            var urlTokens = new UrlTokens(linkUrls, false);
            
            try
            {
                if (! urlTokens.schemes.equal)
                    throw "scheme.notEqual";
                
                if (! urlTokens.hostTokens.slice(1).every(function(sa) { return sa.equal; }))
                    throw "host.notEqual";
                
                if (urlTokens.pathTokens[urlTokens.pathTokens.length - 1].hasNull)
                    throw "path.notEqualCount";
                
                model.linkUrlTokens = urlTokens;
                model.target = ((this.targetCount.image > 0) ? TARGET_IMAGE : TARGET_PAGE);
            }
            catch (e if (typeof(e) == "string"))
            {
                model.error = e;
                return advance("error");
            }
        }
        
        return true;
    },
    
    onThumbnailTreeSelect: function()
    {
        this._updateSelected();
        updateWizardButtons();
    },
    
    onThumbnailTreeKeyUp: function()
    {
        this._setThumbnail(this.tree.currentIndex);
    },
    
    // XXX: mouse scroll event doesn't work properly in ff2: https://bugzilla.mozilla.org/show_bug.cgi?id=352179
    
    onThumbnailTreeMouseMove: function(e)
    {
        var row = {};
        var col = {};
        var childElement = {};
        this.tree.treeBoxObject.getCellAt(e.clientX, e.clientY, row, col, childElement);
        var row = row.value;
        if (row >= 0)
            this._setThumbnail(row);
    },
    
    onThumbnailTreeMouseOut: function(e)
    {
        this._setThumbnail(null);
    },
    
    onThumbnailTargetUpdate: function(thumbnail)
    {
        var row = thumbnail.index;
        var col = this.tree.columns["tcol_thumbnails_target"];
        this.tree.boxObject.invalidateCell(row, col);
        this._targetCount = null;
        updateWizardButtons();
    },
    
    canAdvance: function()
    {
        if (model.selected.length < 2)
            return false;
        var count = this.targetCount;
        if ((count.unknown + count.error) > 0)
            return false;
        if ((count.image + count.page) < model.selected.length)
            return false;
        if (! ((count.image == 0) || (count.page == 0)))
            return false;
        return true;
    },
    
    _updateSelected: function()
    {
        model.selected = [model.thumbnails[row] for each (row in getTreeSelectionIterator(this.tree.view.selection))];
        this._targetCount = null;
        $("lbl_thumbnails_treeStatus").value = _("count", model.selected.length, model.thumbnails.length);
    },
    
    _setThumbnail: function(index)
    {
        if (index == this.thumbnailIndex)
            return;
        
        this.thumbnailIndex = index;
        
        var image = $("image_thumbnail");
        if (index != null)
        {
            var thumbnail = model.thumbnails[index];
            
            var boxComputedStyle = document.defaultView.getComputedStyle($("box_thumbnail"), null);
            var boxWidth = boxComputedStyle.width;
            var boxHeight = boxComputedStyle.height;
            boxWidth = Number(boxWidth.substr(0, boxWidth.length - 2));     // "- 2": strip 'px'
            boxHeight = Number(boxHeight.substr(0, boxHeight.length - 2));
            var thumbnailWidth = thumbnail.image.naturalWidth;
            var thumbnailHeight = thumbnail.image.naturalHeight;
            var factor = Math.min(1, Math.min(boxWidth / thumbnailWidth, boxHeight / thumbnailHeight));
            
            image.src = thumbnail.image.src;
            image.style.width = Math.floor(factor * thumbnailWidth) + "px";
            image.style.height = Math.floor(factor * thumbnailHeight) + "px";
            $("lbl_thumbnailSize").value = _("imageSize", thumbnailWidth, thumbnailHeight);
        }
        else
        {
            image.src = "";
            $("lbl_thumbnailSize").value = "";
        }
    }
}


pages.mode =
{
    tree: null,
    
    get mode()      { return $("rgroup_mode").value; },
    set mode(value) { $("rgroup_mode").value = value; },
    
    get isCreate() { return (this.mode == MODE_CREATE); },
    get isModify() { return (this.mode == MODE_MODIFY); },
    
    get scriptId()      { return Number($("mlist_mode_scripts").value); },
    set scriptId(value) { $("mlist_mode_scripts").value = value; },
    
    get script() { return (this.scriptId ? scriptMgr.getScript(this.scriptId) : null); },
    
    onPageShow: function()
    {
        if (! model.mode)
        {
            $("rgroup_mode").selectedIndex = -1;
            $("col_mode_label").style.width = _("css.labelColumn.width");
            
            this.tree = $("tree_mode");
            this.tree.view = new ThumbnailTreeView(model.selected);
            
            var popup = $("mpopup_mode_scripts");
            while (popup.hasChildNodes())
                popup.removeChild(popup.firstChild);
            
            popup.appendChild(createXulElement("menuitem", {label: "", value: 0}));
            $("mlist_mode_scripts").selectedIndex = 0;
            
            for each (script in model.names)
                popup.appendChild(createXulElement("menuitem", {label: script.name, value: script.id}));
            
            this._updateScript();
        }
    },
    
    onPageAdvanced: function()
    {
        model.mode = this.mode;
        model.scriptId = (model.isModify ? this.scriptId : null);
        if (model.isCreate)
            model.type = ((model.target == TARGET_IMAGE) ? TYPE_DIRECT : null);
        else if (model.isModify)
            model.type = model.script.type;
        return true;
    },
    
    onModeSelect: function()
    {
        this._updateScript();
        updateWizardButtons();
    },
    
    onScriptSelect: function()
    {
        this._updateScript();
        updateWizardButtons();
    },
    
    canAdvance: function()
    {
        if (this.isCreate)
            return true;
        if (this.isModify && this.script)
            switch (this.script.type)
            {
                case TYPE_DIRECT:
                    return (model.target == TARGET_IMAGE);
                case TYPE_REGEXP:
                    return (model.target == TARGET_PAGE);
            }
        return false;
    },
    
    _updateScript: function()
    {
        var script = this.script;
        if (script)
        {
            $("tbox_mode_scriptMatcher").value = script.regexp;
            $("tbox_mode_scriptType").value = _("type." + script.type);
            
            if (script.type == TYPE_REGEXP)
            {
                $("tbox_mode_scriptPatternSource").value = _("patternSource." + script.source);
                $("tbox_mode_scriptPattern").value = script.pattern;
                $("tbox_mode_scriptTemplate").value = script.template;
            }
        }
        
        $("mlist_mode_scripts").parentNode.parentNode.hidden = (! this.isModify);
        $("tbox_mode_scriptMatcher").parentNode.hidden  = (! script);
        $("tbox_mode_scriptType").parentNode.hidden = (! script);
        $("tbox_mode_scriptPatternSource").parentNode.hidden = (! (script && (script.type == TYPE_REGEXP)));
        $("tbox_mode_scriptPattern").parentNode.hidden = (! (script && (script.type == TYPE_REGEXP)));
        $("tbox_mode_scriptTemplate").parentNode.hidden = (! (script && (script.type == TYPE_REGEXP)));
        
    }
}


pages.name =
{
    // TODO: use tbox.defaultValue instead of 0?
    
    tree: null,
    
    get name()      { return $("tbox_name").value; },
    set name(value) { $("tbox_name").value = value; },
    
    onPageShow: function()
    {
        if (! model.name0)
        {
            this.tree = $("tree_name");
            this.tree.view = new ThumbnailTreeView(model.selected);
            
            model.name0 = (model.isModify ? model.script.name : this._resolveName());
            this.name = model.name0;
            
            this._updateButtons();
        }
    },
    
    onPageAdvanced: function()
    {
        model.name = this.name;
        return true;
    },
    
    onNameInput: function()
    {
        this._updateButtons();
        updateWizardButtons();
    },
    
    resetName: function()
    {
        this.name = model.name0;
        this._updateButtons();
        updateWizardButtons();
    },
    
    canAdvance: function()
    {
        return (this.name.length > 0);
    },
    
    _updateButtons: function()
    {
        $("btn_name_reset").disabled = (this.name == model.name0);
    },
    
    _resolveName: function()
    {
        var urlTokens = model.linkUrlTokens;
        var hostToken = (urlTokens.hostTokens[0].equal ? urlTokens.hostTokens[0].strings[0] : null);
        var index = ((hostToken && (! /^www/.test(hostToken))) ? 0 : 1);
        var name = [sa.strings[0] for each (sa in urlTokens.hostTokens.slice(index))].join(".");
        return name;
    }
}


pages.matcher =
{
    tree: null,
    treeView: null,
    
    get matcher() { return $("tbox_matcher").value; },
    set matcher(value)
    {
        $("tbox_matcher").value = value;
        this._matcherRegExp = null;
    },
    
    get matcherRegExp()
    {
        if (! this._matcherRegExp)
            try { this._matcherRegExp = new RegExp("^" + this.matcher + "$"); }
            catch (e) { /* ignore */ }
        return this._matcherRegExp;
    },
    
    onPageShow: function()
    {
        if (! model.matcher0)
        {
            this.tree = $("tree_matcher");
            this.treeView = new ThumbnailTreeView(model.selected);
            this.tree.view = this.treeView;
            
            model.matcher0 = (model.isModify ? model.script.regexp : this._resolveMatcher());
            this.matcher = model.matcher0;
            
            this.onMatcherInput();
        }
    },
    
    onPageAdvanced: function()
    {
        model.matcher = this.matcher;
        if (model.type == TYPE_DIRECT)
            return advance("script");
        else
            return advance("images");
    },
    
    onMatcherInput: function()
    {
        this._matcherRegExp = null;
        if (this.matcherRegExp)
        {
            var matches = [this.matcherRegExp.test(thumbnail.anchor.href) for each (thumbnail in model.selected)];
            this.matchCount = count(matches);
        }
        else
        {
            var matches = [false for each (thumbnail in model.selected)];
            this.matchCount = 0;
        }
        this.treeView.rowStates = matches;
        this._update();
    },
    
    resetMatcher: function()
    {
        this.matcher = model.matcher0;
        this.onMatcherInput();
    },
    
    canAdvance: function()
    {
        return (this.matchCount == model.selected.length);
    },
    
    _update: function()
    {
        $("btn_matcher_reset").disabled = (this.matcher == model.matcher0);
        $("lbl_matcher_treeStatus").value = _("count", this.matchCount, model.selected.length);
        updateWizardButtons();
    },
    
    _resolveMatcher: function()
    {
        return resolveUrlRegExp(model.linkUrlTokens);
    }
}


pages.images =
{
    images: [],
    
    onPageShow: function()
    {
        // TODO: requires (re-)init if model.images = null
        
        var left_rlbox = $("rlbox_thumbnails");
        var deck = $("deck_targets");
        deck.selectedPanel.selectedIndex = -1;
        
        for (var index = 0; index < model.selected.length; index++)
        {
            this.images[index] = null;
            var thumbnail = model.selected[index];
            
            // left richlistbox
            var maxWidth = 128;
            var maxHeight = 128;
            var width = thumbnail.image.naturalWidth;
            var height = thumbnail.image.naturalHeight;
            var factor = Math.min(1, Math.min(maxWidth / width, maxHeight / height));
            var image = createHtmlElement("img", {src: thumbnail.image.src});
            image.style.width = Math.floor(factor * width) + "px";
            image.style.height = Math.floor(factor * height) + "px";
            
            var box = createXulElement("box", null, [image]);
            var rlitem = createXulElement("richlistitem", null, [box]);
            left_rlbox.appendChild(rlitem);
            
            // right richlistbox
            image = createHtmlElement("img");
            box = createXulElement("box", null, [image]);
            rlitem = createXulElement("richlistitem", {"class": "loading page"}, [box]);
            var rlbox = createXulElement("richlistbox", {suppressonselect: true}, [rlitem]);
            deck.appendChild(rlbox);
            rlbox.selectedIndex = -1;
            
            // target loading
            var url = thumbnail.linkUrl;
            var headers = {"Referer": window.arguments[0].location.href};
            var onSuccess = getCallback(this, this._onPageLoadSuccess, index);
            var onError = getCallback(this, this._onPageLoadError, index);
            var timeout = nsPreferences.getIntPref("extensions.thumbsdown.scriptwizard.requestTimeout");
            var resourceLoader = new ResourceLoader("GET", url, headers, onSuccess, onError, timeout);
            resourceLoader.load();
        }
        
        this._updateSelectedCount();
    },
    
    onPageAdvanced: function()
    {
        // TODO: model should have the logic for equality check (if equal -> nop)
        var selectedImages  = [];
        for (var i = 0; i < model.selected.length; i++)
            if (this.images[i])
            {
                model.selected[i].targetImage = this.images[i];
                selectedImages.push(model.selected[i]);
            }
        
        model.images = selectedImages;
        return advance("regexp");
    },
    
    onThumbnailSelect: function()
    {
        this._updateThumbnail();
        this._updateImage();
    },
    
    onImageSelect: function(e)
    {
        var rlbox = e.target;
        var index = rlbox.parentNode.selectedIndex - 1;
        var image = rlbox.selectedItem.firstChild.firstChild;
        this.images[index] = image;
        this._updateImage();
        this._updateSelectedCount();
        updateWizardButtons();
    },
    
    canAdvance: function()
    {
        return (count(this.images) >= 2);
    },
    
    canRewind: function()
    {
        // TODO: regexp.canRewind
        // because this needs a rewrite anyway, no point in implementing this
        return false;
    },
    
    _onPageLoadSuccess: function(loader, index)
    {
        try
        {
            var doc = loader.content;
            
            var deck = $("deck_targets");
            var rlbox = deck.childNodes[index + 1];
            rlbox.removeChild(rlbox.firstChild);
            rlbox.suppressOnSelect = false;
            
            for (var i = 0; i < doc.images.length; i++)
            {
                // FIXME: change page before image loads and  _onImageLoad doesn't exist :)
                var image = createHtmlElement("img", {src: doc.images[i].src, onload: "page._onImageLoad(event)"});
                var box = createXulElement("box", null, [image]);
                var rlitem = createXulElement("richlistitem", {"class": "loading image", value: i}, [box]);
                rlbox.appendChild(rlitem);
                rlbox.setAttribute("onselect", "page.onImageSelect(event)");
                // TODO: rlbox (seltype=single) doesn't support deselect - simulate (de)select using onclick?
            }
        }
        finally
        {
            loader.destroy();
        }
    },
    
    _onPageLoadError: function(loader, index)
    {
        try
        {
            // TODO: page load error
        }
        finally
        {
            loader.destroy();
        }
    },
    
    _onImageLoad: function(e)
    {
        var image = e.target;
        
        var maxWidth = 128;
        var maxHeight = 128;
        var width = image.naturalWidth;
        var height = image.naturalHeight;
        var factor = Math.min(1, Math.min(maxWidth / width, maxHeight / height));
        image.style.width = Math.floor(factor * width) + "px";
        image.style.height = Math.floor(factor * height) + "px";
        
        image.parentNode.parentNode.className = "";
    },
    
    // TODO: onImageError
    
    _updateThumbnail: function()
    {
        var rlbox = $("rlbox_thumbnails");
        var index = rlbox.currentIndex;
        $("deck_targets").selectedIndex = index + 1;
        var thumbnail = model.selected[rlbox.currentIndex];
        $("tbox_images_thumbnail").value = thumbnail.image.src;
        $("tbox_images_link").value = thumbnail.anchor.href;
    },
    
    _updateImage: function()
    {
        var rlbox = $("deck_targets").selectedPanel;
        var rlitem = rlbox.selectedItem;
        if (rlitem)
        {
            var image = rlitem.firstChild.firstChild;
            $("tbox_images_image").value = image.src;
            $("lbl_images_imageSize").value = _("imageSize", image.naturalWidth, image.naturalHeight);
        }
        else
        {
            $("tbox_images_image").value = "";
            $("lbl_images_imageSize").value = "";
        }
    },
    
    _updateSelectedCount: function()
    {
        $("lbl_images_selectedCount").value = _("count", count(this.images), model.selected.length);
    }
}


pages.regexp =
{
    // TODO: pattern.(thumbnail/link/selected)
    // sourceId, sourceName
    
    patternThumbnail0: null,
    patternLink0: null,
    templateThumbnail0: null,
    templateLink0: null,
    matchCount: [0, 0],
    
    get sourceMatchCount() { return this.matchCount[0]; },
    get imagesMatchCount() { return this.matchCount[1]; },
    
    get patternSource() { return Number($("radio_patternSource").value); },
    
    get pattern() { return $("deck_pattern").selectedPanel.value; },
    set pattern(value)
    {
        $("deck_pattern").selectedPanel.value = value;
        this._patternRegExp = null;
    },
    
    get patternThumbnail(value) { return $("tbox_pattern_thumbnail").value; },
    set patternThumbnail(value)
    {
        $("tbox_pattern_thumbnail").value = value;
        this._patternRegExp = null;
    },
    
    get patternLink(value) { return $("tbox_pattern_link").value; },
    set patternLink(value)
    {
        $("tbox_pattern_link").value = value;
        this._patternRegExp = null;
    },
    
    get template()      { return $("deck_template").selectedPanel.value; },
    set template(value) { $("deck_template").selectedPanel.value = value; },
    
    get templateThumbnail()      { return $("tbox_template_thumbnail").value; },
    set templateThumbnail(value) { $("tbox_template_thumbnail").value = value; },
    
    get templateLink()      { return $("tbox_template_link").value; },
    set templateLink(value) { $("tbox_template_link").value = value; },
    
    get patternRegExp()
    {
        if (! this._patternRegExp)
            try { this._patternRegExp = new RegExp("^" + this.pattern + "$"); }
            catch (e) { /* ignore */ }
        return this._patternRegExp;
    },
    
    get getSource()
    {
        if (! this._getSource)
            switch (this.patternSource)
            {
                case TYPE_REGEXP_THUMBNAIL:
                    this._getSource = function(thumbnail) { return thumbnail.image.src; };
                    break;
                case TYPE_REGEXP_LINK:
                    this._getSource = function(thumbnail) { return thumbnail.anchor.href; };
                    break;
            }
        return this._getSource;
    },
    
    onPageShow: function()
    {
        if (model.pattern)
            return;
        
        var lboxThumbnails = $("lbox_regexp_thumbnails");
        var lboxLinks = $("lbox_regexp_links");
        var lboxImages = $("lbox_regexp_images");
        for (var i = 0; i < model.images.length; i++)
        {
            var thumbnail = model.images[i];
            lboxThumbnails.appendItem(thumbnail.image.src);
            lboxLinks.appendItem(thumbnail.anchor.href);
            lboxImages.appendItem(thumbnail.targetImage.src);
        }
        
        var script = (model.isModify ? model.script : null);
        var isOldThumbnail = (script && (script.source == TYPE_REGEXP_THUMBNAIL));
        var isOldLink = (script && (script.source == TYPE_REGEXP_LINK));
        
        this.patternThumbnail0 = (isOldThumbnail ? script.pattern : this._resolvePattern("thumbnail"));
        this.patternLink0 = (isOldLink ? script.pattern : this._resolvePattern("link"));
        this.patternThumbnail = this.patternThumbnail0;
        this.patternLink = this.patternLink0;
        
        this.templateThumbnail0 = (isOldThumbnail ? script.template : this._resolveTemplate("thumbnail"));
        this.templateLink0 = (isOldLink ? script.template : this._resolveTemplate("link"));
        this.templateThumbnail = this.templateThumbnail0;
        this.templateLink = this.templateLink0;
        
        this._update(true, true);
        this._updateButtons();
    },
    
    onPageAdvanced: function()
    {
        model.type = TYPE_REGEXP;
        model.patternSource = this.patternSource;
        model.pattern = this.pattern;
        model.template = this.template;
        return advance("script");
    },
    
    onPatternSourceSelect: function()
    {
        var sourceNames = ["thumbnail", "link"];
        $("deck_pattern").selectedIndex = this.patternSource;
        $("deck_template").selectedIndex = this.patternSource;
        $("deck_sources").selectedIndex = this.patternSource;
        $("lbl_pattern").control = "tbox_pattern_" + sourceNames[this.patternSource];
        $("lbl_template").control = "tbox_template_" + sourceNames[this.patternSource];
        
        this._patternRegExp = null;
        this._getSource = null;
        this._update(true, true);
        this._updateButtons();
        updateWizardButtons();
    },
    
    onPatternInput: function()
    {
        this._patternRegExp = null;
        this._update(true, true);
        this._updateButtons();
        updateWizardButtons();
    },
    
    onTemplateInput: function()
    {
        this._patternRegExp = null;
        this._update(true, true);
        this._updateButtons();
        updateWizardButtons();
    },
    
    resetPattern: function()
    {
        this.pattern = (this.patternSource ? this.patternLink0 : this.patternThumbnail0);
        this._update(true, true);
        this._updateButtons();
        updateWizardButtons();
    },
    
    resetTemplate: function()
    {
        this.template = (this.patternSource ? this.templateLink0 : this.templateThumbnail0);
        this._update(true, true);
        this._updateButtons();
        updateWizardButtons();
    },
    
    canAdvance: function()
    {
        if (this.sourceMatchCount != model.images.length)
            return false;
        if (this.imagesMatchCount != model.images.length)
            return false;
        return true;
    },
    
    canRewind: function()
    {
        // TODO: regexp.canRewind
        // because this needs a rewrite anyway, no point in implementing this
        return false;
    },
    
    _resolvePattern: function(sourceName)
    {
        var urls = [thumbnail[sourceName + "Url"] for each (thumbnail in model.selected)];
        var urlTokens = new UrlTokens(urls);
        return resolveUrlRegExp(urlTokens, true);
    },
    
    _resolveTemplate: function(sourceName)
    {
        var targets = new StringArray([thumbnail.targetImage.src for each (thumbnail in model.images)]);
        var template = targets.prefix;
        targets = targets.substring(targets.prefix.length);
        if (targets.isEmpty)
            return template;
        
        var sources = [((sourceName == "thumbnail") ? thumbnail.image.src : thumbnail.anchor.href) for each (thumbnail in model.images)];
        var pattern = $("tbox_pattern_" + sourceName).value;
        var re = new RegExp("^" + pattern + "$");
        
        var matches = [];
        for (var i = 0; i < model.images.length; i++)
            matches[i] = re.exec(sources[i]);
        
        var saMatches = []
        for (var i = 1; i < matches[0].length; i++)
            saMatches.push(new StringArray([match[i] for each (match in matches)]));
        
        do
        {
            for (var index = 0; index < saMatches.length; index++)
                if (targets.startsWith(saMatches[index]))
                    break;
            if (index == saMatches.length)
                return template + targets.strings[0];
            template += "$" + (index + 1);
            var remainders = [];
            for (var i = 0; i < targets.size; i++)
                remainders[i] = targets.strings[i].substr(saMatches[index].strings[i].length);
            targets = new StringArray(remainders);
            template += targets.prefix;
            targets = targets.substring(targets.prefix.length);
        }
        while (! targets.isEmpty)
        
        return template;
    },
    
    _updateButtons: function()
    {
        $("btn_pattern_reset").disabled = (this.pattern == (this.patternSource ? this.patternLink0 : this.patternThumbnail0));
        $("btn_template_reset").disabled = (this.template == (this.patternSource ? this.templateLink0 : this.templateThumbnail0));
    },
    
    _update: function(updateSource, updateImages)
    {
        var update = [updateSource, updateImages];
        var lboxes = [$("deck_sources").selectedPanel, $("lbox_regexp_images")];
        
        if (this.patternRegExp)
        {
            var re = this.patternRegExp;
            var template = this.template;
            var getSource = this.getSource;
            var matches = [[], []];
            for (var i = 0; i < model.images.length; i++)
            {
                var thumbnail = model.images[i];
                var values = [getSource(thumbnail), thumbnail.targetImage.src];
                
                if (updateSource)
                    matches[0].push(re.test(values[0]));
                if (updateImages)
                    matches[1].push(values[0].replace(re, template) == values[1]);
            }
            
            for (var i = 0; i < update.length; i++)
                this.matchCount[i] = count(matches[i]);
        }
        else
        {
            var matches = arrayWithValue(model.images.length, false);
            matches = [matches, matches];
            for (var i = 0; i < update.length; i++)
                this.matchCount[i] = 0;
        }
        
        for (var i = 0; i < update.length; i++)
            if (update[i])
            {
                for (var k = 0; k < model.images.length; k++)
                   lboxes[i].getItemAtIndex(k).className = (matches[i][k] ? "pass" : "fail");
                
                var index = matches[i].indexOf(false);
                if (index > 0)
                   lboxes[i].ensureIndexIsVisible(index);
            }
    }
}


pages.error =
{
    tree: null,
    
    onPageShow: function()
    {
        $("lbl_error_message").textContent = _("error." + model.error);
        
        this.tree = $("tree_error");
        this.tree.view = new ThumbnailTreeView(model.selected);
    },
    
    onPageRewound: function()
    {
        model.error = null;
        return true;
    },
    
    canAdvance: function()
    {
        return false;
    }
}


pages.script =
{
    onPageShow: function()
    {
        var script = model;
        var msg = _("finish.save");
        
        if (model.match)
        {
            script = model.match;
            msg = _("finish.match");
        }
            
        $("lbl_script_msg").textContent = msg;
        $("tbox_script_name").value = script.name;
        $("tbox_script_matcher").value = script.matcher;
        $("tbox_script_type").value = _("type." + script.type);
        
        // FIXME: if script exists, this doesn't quite work
        if (script.isRegExp)
        {
            $("tbox_script_patternSource").value = _("patternSource." + script.patternSource);
            $("tbox_script_pattern").value = script.pattern;
            $("tbox_script_template").value = script.template;
        }
        
        $("tbox_script_patternSource").parentNode.hidden = (! script.isRegExp);
        $("tbox_script_pattern").parentNode.hidden = (! script.isRegExp);
        $("tbox_script_template").parentNode.hidden = (! script.isRegExp);
    }
}


function onWizardFinish()
{
    if (model.match)
        return true;
    
    var script = (model.isCreate ? {} : model.script);
    script.name = model.name;
    script.regexp = model.matcher;
    script.type = model.type;
    script.source = model.patternSource;
    script.pattern = model.pattern;
    script.template = model.template;
    if (model.isCreate)
        scriptMgr.addScript(script);
    else
        scriptMgr.setScript(script);
    
    return true;
}


function UrlTokens(urls)
{
    this.urls = urls;
    
    this.schemes = new StringArray([url.scheme for each (url in urls)]);
    this.hosts   = new StringArray([url.host   for each (url in urls)]);
    this.paths   = new StringArray([url.filePath for each (url in urls)]);
    this.queries = new StringArray([url.query  for each (url in urls)]);
    
    this.hostTokens = this.hosts.tokenize(".");
    this.pathTokens = this.paths.tokenize("/");
    this.queryTokens = this.queries.tokenize("&");
}


function resolveUrlRegExp(urlTokens, capture)
{
    var sentinels;
    var tokens;
    
    // scheme
    var scheme = resolveRegExp(urlTokens.schemes);
    
    // host
    sentinels = arrayWithValue(urlTokens.hostTokens.length, ".");
    sentinels[sentinels.length - 1] = "/";
    tokens = [];
    for (var i = 0; i < urlTokens.hostTokens.length; i++)
        tokens.push(resolveRegExp(urlTokens.hostTokens[i], capture, sentinels[i]));
    var host = tokens.join(escapeRegExp("."));
    
    // path
    sentinels = arrayWithValue(urlTokens.pathTokens.length, "/");
    sentinels[sentinels.length - 1] = (urlTokens.queries.isEmpty ? null : '?');
    tokens = [];
    for (var i = 0; i < urlTokens.pathTokens.length; i++)
        tokens.push(resolveRegExp(urlTokens.pathTokens[i], capture, sentinels[i]));
    var path = tokens.join(escapeRegExp("/"));
    
    // query
    sentinels = arrayWithValue(urlTokens.queryTokens.length, "&");
    sentinels[sentinels.length - 1] = null;
    tokens = [];
    for (var i = 0; i < urlTokens.queryTokens.length; i++)
        tokens.push(resolveRegExp(urlTokens.queryTokens[i], capture, sentinels[i]));
    var query = tokens.join(escapeRegExp("&"));
    if (query)
        query = escapeRegExp("?") + query;
    
    return scheme + "://" + host + path + query;
}

function resolveRegExp(sa, capture, sentinel)
{
    if (sa.equal)
        return escapeRegExp(sa.strings[0]);
    
    function resolve(sa, sentinel)
    {
        var group = null;
        var chars = "";
        var pattern = null;
        var multiplier = ((sa.length.min == 0) ? "*" : "+");
        
        var i = 0, k = 0;
        do
        {
            var str = sa.strings[i];
            var c = str.charAt(k);
            
            if ((! group) && /\d/.test(c))
                group = "\\d";
            else if ((group != "\\w") && /\w/.test(c))
                group = "\\w";
            else if (chars.length < 1)
                chars = ((c == "-") ? c + chars : chars + c);
            else if (! pattern)
                pattern = (sentinel ? "[^" + sentinel + "]" : ".");
            else
                throw new Error("unreachable code");
            
            var p = (pattern ? pattern : (chars ? "[" + escapeRegExp(chars) + group + "]" : group)) + multiplier;
            var re = new RegExp("^" + p);
            
            for (i = 0; i < sa.strings.length; i++)
            {
                var match = re.exec(str);
                if (match[0].length != str.length)
                {
                    k = match[0].length;
                    break;
                }
            }
        }
        while (i < sa.strings.length);
        
        return p;
    }
    
    var prefix = escapeRegExp(sa.prefix);
    var middle = resolve(sa.middle(), sentinel);
    var suffix = escapeRegExp(sa.suffix);
    
    if (capture)
        middle = "(" + middle +")";
    
    return prefix + middle + suffix;
}


function ThumbnailTreeView(thumbnails)
{
    this.thumbnails = thumbnails;
}

ThumbnailTreeView.prototype =
{
    get rowStates() { return this._rowStates; },
    set rowStates(value)
    {
        this.treeBoxObject.beginUpdateBatch();
        var firstFail = null;
        if (this._rowStates !== undefined)
            for (var row = 0; row < value.length; row++)
            {
                if (value[row] != this._rowStates[row])
                    this.treeBoxObject.invalidateRow(row);
                if ((firstFail == null) && (! value[row]))
                    firstFail = row;
            }
        if (firstFail != null)
            this.treeBoxObject.ensureRowIsVisible(firstFail);
        this._rowStates = value;
        this.treeBoxObject.endUpdateBatch();
    },
    
    get rowCount() { return this.thumbnails.length; },
    
    setTree: function(treeBoxObject)
    {
        this.treeBoxObject = treeBoxObject;
    },
    
    getCellText: function(row, col)
    {
        var thumbnail = this.thumbnails[row];
        switch (col.id)
        {
            case "tcol_thumbnails_link":
            case "tcol_mode_link":
            case "tcol_name_link":
            case "tcol_matcher_link":
            case "tcol_error_link":
                return thumbnail.anchor.href;
            
            case "tcol_thumbnails_target":
                var target = thumbnail.target;
                if (! target)
                    return "";
                switch (target.type)
                {
                    case "image":   return _("target.image");
                    case "page":    return _("target.page");
                    case "unknown": return (target.contentType ? _("target.unknown", (target.mimeType ? target.mimeType : target.contentType)) : _("target.noContentType"));
                    case "error":   return _("target.error", (target.status ? target.status : -1));
                    case "timeout": return _("target.timeout");
                }
                throw new Error("unknown target: " + target.type);
        }
        throw new Error("unknown column: " + col.id);
    },
    
    getRowProperties: function(row, props)
    {
        if (this.rowStates !== undefined)
            props.AppendElement(this._getAtom(this.rowStates[row] ? "pass" : "fail"));
    },
    
    isContainer: function(row) { return false; },
    isSeparator: function(row) { return false; },
    isSorted: function() { return false; },
    getLevel: function(row) { return 0; },
    getImageSrc: function(row, col) { return null; },
    getCellProperties: function(row, col, props) {},
    getColumnProperties: function(col, props) {},
    
    // http://www.xulplanet.com/tutorials/xultu/treestyle.html
    _atoms: {},
	_getAtom: function(a)
    {
		if (! (a in this._atoms))
			this._atoms[a] = atomSvc.getAtom(a);
		return this._atoms[a];
	}
}
